from neo4j_uploader._logger import ModuleLogger
from neo4j_uploader._queries import specification_queries
from neo4j_uploader._n4j import reset, upload_query, validate_credentials
from neo4j_uploader._upload_utils import upload_nodes, upload_relationships
from neo4j_uploader.models import UploadResult, Neo4jConfig, GraphData
from neo4j_uploader.errors import InvalidCredentialsError, InvalidPayloadError
from neo4j_uploader._conversions import convert_legacy_node_records, convert_legacy_relationship_records
from timeit import default_timer as timer
from warnings import warn
import json

# Specify Google doctstring type for pdoc auto doc generation
__docformat__ = "google"

def start_logging():
    """
    Enables logging from this module. Log level matches the existing log level of the calling module.
    """
    logger = ModuleLogger()
    logger.is_enabled = True
    logger.info("Neo4j Uploader logging enabled")

def stop_logging():
    """
    Surpresses logging from this module.
    """
    ModuleLogger().info(f'Discontinuing logging')
    ModuleLogger().is_enabled = False


def batch_upload(
        config: dict | Neo4jConfig,
        data : dict | GraphData,
    ) -> UploadResult:
    """Uploads a dictionary containing nodes, relationships, and target Neo4j database information. The schema for nodes and relationships is more flexible and comprehensive than the schema for the earlier upload function.

    Args:
        config (dict or Neo4jConfig): A Neo4jConfig object or dict that can be converted to a Neo4jConfig object for defining target Neo4j database and credentials for upload.
        data (dict or GraphData): A GraphData object or a dict that can be converted to a GraphData object with specifications for nodes and relationships to upload

        
    Returns:
        UploadResult: Result object containing information regarding a successful or unsuccessful upload.

    Raises:
        neo4j.exceptions: A Neo4j exception if credentials are invalid or database can not be accessed.
        InvalidCredentialsError: If credentials are missing or malformed.
        InvalidPayloadError: If payload schema is missing or unsupported.
    """
     
    try:
        cdata = Neo4jConfig.model_validate(config)
    except Exception as e:
        raise InvalidCredentialsError(e)
    
    # Will raise a neo4j.exception if credentials failed or database can not be accessed
    validate_credentials((cdata.neo4j_uri, cdata.neo4j_user, cdata.neo4j_password))
    
    try:
        gdata = GraphData.model_validate(data)
    except Exception as e:
        raise InvalidPayloadError(e)


    # Start clock for tracking processing time
    start = timer()
    total_nodes_created = 0
    total_relationships_created = 0
    total_properties_set = 0

    # Get list of tuples containing queries and accompanying params for driver execution
    query_params = specification_queries(gdata.nodes, cdata)
    query_params.extend(specification_queries(gdata.relationships, cdata))

    for qp in query_params:
        # Run queries and retrieve summary of upload
        summary = upload_query(
            creds = (cdata.neo4j_uri, cdata.neo4j_user, cdata.neo4j_password),
            query = qp[0],
            params = qp[1],
            database = cdata.neo4j_database
        )
        
        # Sample summary result
        # {'metadata': {'query': '<query>', 'parameters': {}, 'query_type': 'w', 'plan': None, 'profile': None, 'notifications': None, 'counters': {'_contains_updates': True, 'labels_added': 17, 'nodes_created': 17, 'properties_set': 78}, 'result_available_after': 73, 'result_consumed_after': 0}

        # {'metadata': {'query': "<rel_upload_query>", 'parameters': {}, 'query_type': 'w', 'plan': None, 'profile': None, 'notifications': None, 'counters': {'_contains_updates': True, 'relationships_created': 1, 'properties_set': 2}, 'result_available_after': 209, 'result_consumed_after': 0}

        # Sum up total nodes, relationshipts and props set
        try:
            props = summary.counters.properties_set
            total_properties_set += props
        except Exception as _:
            ModuleLogger().debug('No properties set in summary: {summary}')

        try:
            nodes = summary.counters.nodes_created
            total_nodes_created += nodes
        except Exception as _:
            pass

        try:
            relationships = summary.counters.relationships_created
            total_relationships_created += relationships
        except Exception as _:
            pass

    stop = timer()
    time_to_complete = round((stop - start), 4)

    return UploadResult(
        was_successful = True,
        error_message = None,
        seconds_to_complete = time_to_complete,
        nodes_created = total_nodes_created,
        relationships_created = total_relationships_created,
        properties_set = total_properties_set
    )

def upload(
    neo4j_creds:(str, str, str), 
    data: str | dict,
    node_key : str = "_uid",
    dedupe_nodes : bool = True,
    dedupe_relationships : bool = True,
    should_overwrite: bool = False,
    database_name: str = 'neo4j',
    max_batch_size: int = 500,
    )-> (float, int, int, int):
    """
    Uploads a dictionary of simple node and relationship records to a target Neo4j instance specified in the arguments.

    Args:
        neo4j_creds: Tuple containing the hostname, username, password, and optionally a database name of the target Neo4j instance. The host name should contain only the database name and not the protocol. For example, if the host name is 'neo4j+s://<unique_db_id>.databases.neo4j.io', the host string to use is '<unique_db_id>.databases.neo4j.io'. The default database name is 'neo4j'.

        data: A .json string or dictionary of records to upload. The dictionary keys must contain a 'nodes' and 'relationships' key. The value of which should be a list of dictionaries, each of these dictionaries contain the property keys and values for the nodes and relationships to be uploaded, respectively.

        node_key: The key in the dictionary that contains the unique identifier for the node. Relationship generation will also use this to find the from and to Nodes it connects to. Default is '_uid'.

        dedupe_nodes: Should nodes only be created once. False means a new node will always be created. True means if an existing node exists, only the properties will be updated. Default True.

        dedupe_relationships: Should relationships only create 1 of a given relationship between the same from and to node. False means a new relationship will always be created. True means if an existing relationship exists between the target nodes, only the properties will be updated. If no prior relationship, a new one will be created. Default True.

        should_overwrite: A boolean indicating whether the upload should overwrite existing data. If set to True, the upload will delete all existing nodes and relationships before uploading. Default is False.

        database_name: String name of target Neo4j database.

        max_batch_size: Integer maximum number of nodes or relationships to upload in a single Cypher batch. Default 500.
    
    Returns:
        Tuple of result data: float of time to complete, int of nodes created, int of relationships created, int of total node and relationship properties set.
    
    Raises:
        Exceptions if data is not in the correct format or if the upload ungracefully fails.
    """

    # Convert to dictionary if data is string
    if isinstance(data, str) is True:
        try:
            data = json.loads(data)
        except Exception as e:
            raise Exception(f'Input data string not a valid JSON format: {e}')
    
    if data is None or len(data) == 0:
        raise Exception(f'data payload is empty or an invalid format')

    simple_nodes = data.get('nodes', None)
    simple_rels = data.get('relationships', None)

    nodes = convert_legacy_node_records(simple_nodes, dedupe_nodes, node_key)

    rels = convert_legacy_relationship_records(simple_rels, dedupe_relationships, node_key)

    uri, user, password = neo4j_creds

    config = Neo4jConfig(
        neo4j_uri = uri,
        neo4j_user = user,
        neo4j_password = password,
        neo4j_database = database_name,
        max_batch_size = max_batch_size,
        overwrite = should_overwrite
    )

    return batch_upload(
        config = config,
        data = {
            "nodes": nodes,
            "relationships": rels
        }
    )

def clear_db(creds: (str, str, str), database: str):
    """Deletes all existing nodes and relationships in a target Neo4j database.

    Args:
        creds (str, str, str): Neo4j URI, username, and password.
        database (str): Target Neo4j database.

    Returns:
        summary (neo4j.ResultSummary): Result summary of the operation. See https://neo4j.com/docs/api/python-driver/current/api.html#resultsummary for more info.
    """
    return reset(creds, database)


import logging

class ModuleLogger(object):
    '''
    @private
    Used to pass logging information to the calling module.
    '''

    is_enabled : bool = False
    _logger = None

    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(ModuleLogger, cls).__new__(cls)
        return cls.instance

    def logger(self):
        if self.is_enabled is False:
            return EmptyLogger()
        
        if self._logger is None:
            self._logger = logging.getLogger(__name__)
            # For custom formatting - This will cause duplicate logs for each message
            # FORMAT = "[%(asctime)s: %(filename)s: %(lineno)s - %(funcName)20s()] %(message)s"
            # formatter = logging.Formatter(FORMAT)
            # mh = logging.StreamHandler()
            # mh.setFormatter(formatter)
            # self._logger.addHandler(mh)
            self._logger.addHandler(logging.NullHandler())
        
        return self._logger

    def notset(self, arg: str):
        if self.is_enabled:
            return self.logger().notset(arg)
        pass

    def debug(self, arg: str):
        if self.is_enabled:
            return self.logger().debug(arg)
        pass
  
    def info(self, arg: str):
        if self.is_enabled:
            return self.logger().info(arg)
        pass

    def warning(self, arg: str):
        if self.is_enabled:
            return self.logger().warning(arg)
        pass

    def error(self, arg: str):
        if self.is_enabled:
            return self.logger().error(arg)
        pass

    def critical(self, arg: str):
        if self.is_enabled:
            return self.logger().critical(arg)
        pass

class EmptyLogger:
    '''
    @private
    '''

    def notset(self, arg:str):
        pass
    def debug(self, arg: str):
        pass
    def info(self, arg: str):
        pass
    def warning(self, arg: str):
        pass
    def error(self, arg: str):
        pass
    def critical(self, arg: str):
        pass


from neo4j_uploader._n4j import execute_query
from neo4j_uploader._logger import ModuleLogger

# Legacy functions

def prop_subquery(
        record: dict, 
        suffix : str = "", 
        exclude_keys: list[str] = []
        )-> (str, dict):
    """
    Generates a Cypher substring statement to set properties for a record, excluding given keys

    Args:
        record: Dict of Node or Relatioship properties to convert

        suffix: String suffix to make parameter values unique

        exclude_keys: List of dictionary key values to ignore from substring generation

    
    Returns:
        A tuple containing the substring and a dict of parameters
    """

    # Validate
    if isinstance(record, dict) == False:
        return ("", {})
    if len(record) == 0:
        return ("", {})
    
    # Embed any prop data within brackets { }
    params = {}
    query = " {"

    filtered_keys = [key for key in record.keys() if key not in exclude_keys]
    sorted_keys = sorted(list(filtered_keys))

    for idx, a_key in enumerate(sorted_keys):
        value = record[a_key]

        # Do not set properties with a None/Null/Empty value
        if value is None:
            continue
        if isinstance(value, str):
            if value.lower() == "none":
                continue
            if value.lower() == "null":
                continue
            if value.lower() == "empty":
                continue
            if value.lower() == "":
                continue

        # Prefix multiple items in Cypgher with comma
        if idx!= 0:
            query += ", "

        # Mark value for parameterization and add to params return dict
        param_key = f'{a_key}_{suffix}'
        params[param_key] = value

        # Cypher string query requires { } to designate parameter values
        query += f'`{a_key}`:${param_key}'

    # Close out query
    query += "}"

    return query, params

class HashableDict:
    def __init__(self, dictionary):
        self.dictionary = dictionary

    def __hash__(self):
        return hash(frozenset(self.dictionary.items()))

# Legacy
def with_node_elements(
    nodes: list[dict],
    node_key: str,
    dedupe : bool = True
) -> (str, dict):
    """
    Returns elements to be added into a batch node creation query.

    Args:
        label: The node label.

        nodes: A list of dictionary records for each node to process.

        nodes_key: The property key that uniquely identifies Nodes. 

        dedupe: Should duplicate nodes be purged from final results. Default True.
    
    Returns:
        A tuple of the query substring and parameters dictionary
    """
    results_list = []
    results_params = {}

    if dedupe == True:
        nodes = [dict(t) for t in {tuple(n.items()) for n in nodes}]

    nodes = sorted(nodes, key=lambda x: x[node_key])

    for idx, node in enumerate(nodes):

        # For helping id params later
        suffix = f"n{idx}"

        # string and params for props
        props_string, params = prop_subquery(
            record = node,
            suffix=suffix,
            exclude_keys=[])
        
        # Only processing batches of properties for Nodes. Still recommended way to handle batches vs. series of MATCH or CREATE statements
        with_element = f"{props_string}"
        results_list.append(with_element)
        results_params = results_params | params

    return results_list, results_params

# Legacy
def upload_node_records_query(
    label: str,
    nodes: list[dict],
    node_key: str = "_uid",
    dedupe : bool = True
    )->(str, dict):
    """
    Generate Cypher Node update query.

    Args:
        label: Label of Nodes to generate

        nodes: A list of dictionaries representing Node properties

        dedupe: Remove duplicate entries. Default True.
    
    Returns:
        A tuple with the full Cypher query statment and dictionary of parameters to pass to the Neo4j driver

    Raises:
        Exceptions if data is not in the correct format or if the upload fails.
    """
        
    if nodes is None:
        return ("", {})
    if len(nodes) == 0:
        return ("", {})
    
    ModuleLogger().debug(f'Starting to process node labeled: {label} ...')

    query = ""

    with_elements, params = with_node_elements(
        nodes,
        node_key,
        dedupe
    )

    if len(with_elements) is None:
        ModuleLogger().warning(f'Could not process nodes labeled {label}. Check if data exsists and matches expected schema')
        return ("", {})

    with_elements_str = ",".join(with_elements)

    # Assemble final query
    query = f"""WITH [{with_elements_str}] AS node_data\nUNWIND node_data AS node"""
    
    if dedupe == True:
        query += f"\nMERGE (n:`{label}` {{`{node_key}`:node.`{node_key}`}})"
    else:
        query += f"\nCREATE (n:`{label}`)"
    query += "\nSET n += node"

    return query, params

# Legacy
def upload_nodes(
    neo4j_creds:(str, str, str),
    nodes: dict,
    node_key: str = "_uid",
    database : str = "neo4j",
    dedupe : bool = True,
    max_batch_size: int = 500
)-> (int, int):
    """
    Uploads a list of dictionary objects as nodes.

    Args:
        neo4j_credits: Tuple containing the hostname, username, and password of the target Neo4j instance

        nodes: A dictionary of objects to upload. Each key is a unique node label and contains a list of records as dictionary objects.

        node_key: String property key that uniquely identifies each node. Default '_uid'.

        database: Name of Neo4j database to connect to. Default 'neo4j'.

        dedupe: Remove duplicate entries. Default True

        max_batched_size: Maximum number of nodes to upload in a single batch. Default 500.
    
    Returns:
        A tuple containing the number of nodes created and properties set.

    Raises:
        Exceptions if data is not in the correct format or if the upload fails.
    """
    if nodes is None:
        return 0, 0
    if len(nodes) == 0:
        return 0, 0
    
    # For reporting
    nodes_created = 0
    props_set = 0

    query = """"""

    ModuleLogger().debug(f'Uploading node records: {nodes}')

    # Sort Nodes to guarantee consistent output
    filtered_keys = [key for key in nodes.keys()]
    sorted_keys = sorted(list(filtered_keys))

    # NOTE: Process each Node label separately so as to not timeout connection with db. Somewhere between 100-1000 Nodes.
    for node_label in sorted_keys:

        total_nodes_list = nodes.get(node_label, None)

        # TODO: Make a calculation of node * properties for a better batch calculation

        # Break relationships into batches of 500
        b = max_batch_size
        chunked_nodes_list = [total_nodes_list[i * b:(i + 1) * b] for i in range((len(total_nodes_list) + b - 1) // b )]  

        for nodes_list in chunked_nodes_list:
            if nodes_list is None:
                ModuleLogger().error(f'No values for node label {node_label} found. Skipping.')
                continue

            # Process all similar labeled nodes together
            nodes_query, nodes_params = upload_node_records_query(
                node_label, 
                nodes_list, 
                dedupe=dedupe, 
                node_key=node_key)

            ModuleLogger().debug(f'upload nodes query: {query}')

            records, summary, keys = execute_query(
                neo4j_creds, 
                nodes_query, 
                params=nodes_params, 
                database=database)

            # Sample summary
            # {'metadata': {'query': '<query>', 'parameters': {}, 'query_type': 'w', 'plan': None, 'profile': None, 'notifications': None, 'counters': {'_contains_updates': True, 'labels_added': 17, 'nodes_created': 17, 'properties_set': 78}, 'result_available_after': 73, 'result_consumed_after': 0}
            try:
                created = summary.counters.nodes_created
                nodes_created += created
            except Exception as e:
                ModuleLogger().debug(f'No nodes labeled: {node_label} created')

            try:
                props = summary.counters.properties_set
                props_set += props
            except Exception as e:
                ModuleLogger().debug(f'No node properties for nodes labeled: {node_label} created')
            
            # ModuleLogger().info(f'Results from upload nodes: \n\tRecords: {records}\n\tSummary: {summary.__dict__}\n\tKeys: {keys}')
    
    return nodes_created, props_set

def target_key(
        record: dict,
        prefix: str,
        override_suffix: str) -> str:
    
    # Original implementation - now takes presidence with multiple options
    explicit_key = f"{prefix}{override_suffix}"
    explicit_value = record.get(explicit_key, None)
    if explicit_value is not None:
        return explicit_key
    
    # find all matching keys
    keys = record.keys()
    filtered_keys = [k for k in keys if k.startswith(prefix)]

    if len(filtered_keys) == 0:
        # No from key found!
        return None
    
    # Take first if multiple matches
    key = sorted(filtered_keys)[0]
    return key

def to_key(
        record: dict,
        node_key: str) -> str:
    return target_key(record, "_to_", node_key)

def from_key(
        record: dict,
        node_key: str) -> str:
    return target_key(record, "_from_", node_key)

# Legacy
def with_relationship_elements(
    type: str,
    relationships: list[dict],
    nodes_key: str = "_uid",
    dedupe: bool = True
    ) -> (str, dict):
    """
    Returns elements to be added into a batch relationship creation query.

    Args:
        type: The relationship type.

        relationships: A list of dictionary records for each relationship property.

        nodes_key: The property key that uniquely identifies Nodes. 

        dedupe: Should duplicate relationships be purged from final results. Default True.
    
    Returns:
        A tuple of the query substring and parameters dictionary
    """
    # TODO: Possible cypher injection entry point?

    results_list = []
    results_params = {}

    if dedupe == True:
        relationships = [dict(t) for t in {tuple(r.items()) for r in relationships}]

    # Force sort bc list is not being consistent
    # Little complicated looking because the from and to node id keys are now dynamic, and not necessarily the same for all relationships
    relationships = sorted(relationships, key=lambda x: (x[from_key(x, nodes_key)], x[to_key(x, nodes_key)]))
    
    for idx, rel in enumerate(relationships):

        suffix = f"r{idx}"

        # Get unique key-value of from and to nodes
        from_node_key = from_key(rel, nodes_key)
        to_node_key = to_key(rel, nodes_key)
        from_node_value = rel[from_node_key]
        to_node_value = rel[to_node_key]

        if from_node_value is None:
            ModuleLogger().warning(f'{type} Relationship from node missing target value for key: {from_key}')
            continue
        if to_node_value is None:
            ModuleLogger().warning(f'{type} Relationship to node missing target value for key: {to_key}')
            continue


        # Validate we from and to nodes to work with
        if from_node_value is None:
            ModuleLogger().warning(f'{type} Relationship missing {from_key} property. Skipping relationship {rel}')
            continue
        if to_node_value is None:
            ModuleLogger().warning(f'{type} Relationship missing {to_key} property. Skipping relationship {rel}')
            continue

        # string and params for props
        props_string, params = prop_subquery(
            rel,
            suffix=suffix,
            exclude_keys=[from_node_key, to_node_key])

        # Add from and to node identifiers as params
        from_node_key_param = f"{from_node_key}_{suffix}"
        to_node_key_param = f"{to_node_key}_{suffix}"

        params[from_node_key_param] = from_node_value
        params[to_node_key_param] = to_node_value

        with_element = f"[${from_node_key_param},${to_node_key_param},{props_string}]"
        results_list.append(with_element)
        results_params = results_params | params
    
    return results_list, results_params


# Legacy
def upload_relationship_records_query(
        type: str,
        relationships: list[dict],
        nodes_key: str,
        dedupe: bool = True,
) -> (str, dict):

        if relationships is None:
            return ("", {})
        if len(relationships) == 0:
            return ("", {})
        
        ModuleLogger().debug(f'Starting to process relationships type: {type} ...')

        # Process all similar labeled nodes together
        with_elements, params = with_relationship_elements(
            type,
            relationships, 
            nodes_key=nodes_key, 
            dedupe=dedupe)

        if len(with_elements) is None:
            ModuleLogger().warning(f'Could not process relationships type {type}. Check if data exsists and matches expected schema')
            return ("", {})
        
        with_elements_str = ",".join(with_elements)

        # Assemble final query
        rel_upload_query = f"""WITH [{with_elements_str}] AS from_to_data\nUNWIND from_to_data AS tuple\nMATCH (fromNode {{`{nodes_key}`:tuple[0]}})\nMATCH (toNode {{`{nodes_key}`:tuple[1]}})"""

        # Merge only updates, creates new if not already existent. Create ALWAYS creates a new relationship
        if dedupe == True:
            rel_upload_query += f"\nMERGE (fromNode)-[r:`{type}`]->(toNode)"
        else:
            rel_upload_query += f"\nCREATE (fromNode)-[r:`{type}`]->(toNode)"
        
        # Update Relationship properties if any
        rel_upload_query +=f"\nSET r += tuple[2]"

        return rel_upload_query, params

# Legacy
def upload_relationships(
    neo4j_creds:(str, str, str),
    relationships: dict,
    nodes_key: str = "_uid",
    dedupe : bool = True,
    database: str = "neo4j",
    max_batch_size: int = 500
)-> (int, int):
    """
    Uploads a list of dictionary objects as relationships.

    Args:
        neo4j_credits: Tuple containing the hostname, username, and password of the target Neo4j instance

        nodes: A dictionary of objects to upload. Each key is a unique relationship type and contains a list of records as dictionary objects.

        nodes_key: The property key that uniquely identifies Nodes.

        dedupe: False means a new relationship will always be created for the from and to nodes. True if existing relationships should only be updated. Note that if several relationships already exist, all matching relationships will get their properties updated. Default True.

        database: String name of target Neo4j database

        max_batch_size: Integer maximum number of relationships to upload in a single Cypher batch. Default 500.
    
    Returns:
        A tuple of relationships created, properties set

    Raises:
        Exceptions if data is not in the correct format or if the upload ungracefully fails.
    """

    # Final query needs to look something like this
    # WITH [['1202109692044412','1204806322568817', {name:"fish"}],['1202109692044411','test', {}]] AS from_to_id_pairs

    # UNWIND from_to_id_pairs as pair

    # MATCH (fromNode { `gid`: pair[0]})
    # MATCH (toNode { `gid`: pair[1]})

    # CREATE (fromNode)-[r:`WITHIN`]->(toNode)
    # SET r = pair[2]


    # Validate
    if relationships is None:
        return 0,0
    if len(relationships) == 0:
        return 0,0
    
    ModuleLogger().debug(f'upload relationships source data: {relationships}')

    # Upload counts
    relationships_created = 0
    props_set = 0

    # Sort so we get a consistent output
    filtered_keys = [key for key in relationships.keys()]
    sorted_keys = sorted(list(filtered_keys))

    # NOTE: Need to process each relationship type separately as batching mixed relationships fails
    for rel_type in sorted_keys:


        total_rel_list = relationships[rel_type]

        # Break relationships into batches
        b = max_batch_size
        chunked_rel_list = [total_rel_list[i * b:(i + 1) * b] for i in range((len(total_rel_list) + b - 1) // b )]  

        for rel_list in chunked_rel_list:

            rel_query, rel_params = upload_relationship_records_query(
                type=rel_type,
                relationships=rel_list,
                nodes_key=nodes_key,
                dedupe=dedupe)

            records, summary, keys = execute_query(
                neo4j_creds, 
                rel_query,params=rel_params, 
                database=database)

            # Sample summary result
            # {'metadata': {'query': "<rel_upload_query>", 'parameters': {}, 'query_type': 'w', 'plan': None, 'profile': None, 'notifications': None, 'counters': {'_contains_updates': True, 'relationships_created': 1, 'properties_set': 2}, 'result_available_after': 209, 'result_consumed_after': 0}

            # Can we have optionals yet?
            try:
                created = summary.counters.relationships_created
                relationships_created += created
            except Exception as e:
                ModuleLogger().debug(f'No relationship type: {rel_type} created')

            try:
                props = summary.counters.properties_set
                props_set += props
            except Exception as _:
                ModuleLogger().debug(f'No node properties for relationship type: {rel_type} created')

            # ModuleLogger().info(f'Results from uploading relationships type: {rel_type}: \n\tRecords: {records}\n\tSummary: {summary.__dict__}\n\tKeys: {keys}')

    return relationships_created, props_set



class InvalidCredentialsError(Exception):
    """Error if Neo4j credentials are invalid"""
    pass

class InvalidPayloadError(Exception):
    """Error if json payload schema is unrecognized"""
    pass


import asyncio
from neo4j import GraphDatabase
from neo4j_uploader._logger import ModuleLogger

def validate_credentials(creds: (str, str, str)):
    host, user, password = creds
    with GraphDatabase.driver(host, auth=(user, password)) as driver:
        driver.verify_connectivity()

def upload_query(creds: (str, str, str), query, params={}, database: str = "neo4j"):
    host, user, password = creds
    with GraphDatabase.driver(host, auth=(user, password)) as driver:
        _, summary, _ = driver.execute_query(query, params, database=database) 
        return summary

def execute_query(creds: (str, str, str), query, params={}, database: str = "neo4j"):
    host, user, password = creds
    ModuleLogger().debug(f'Using host: {host}, user: {user} to execute query: {query}')
    # Returns a tuple of records, summary, keys
    with GraphDatabase.driver(host, auth=(user, password)) as driver:
        return driver.execute_query(query, params, database=database)

def drop_constraints(creds: (str, str, str), database: str = "neo4j"):
    query = 'SHOW CONSTRAINTS'
    result = execute_query(creds, query, database=database)

    ModuleLogger().info(f"Drop constraints results: {result}")

    # Have to make a drop constraint request for each individually!
    for record in result.records:
        constraint_name = record.get("name", None)
        if constraint_name is not None:
            drop_query = f"DROP CONSTRAINT {constraint_name}"
            drop_result = execute_query(creds, drop_query, database=database)

        ModuleLogger().info(f"Drop constraint {constraint_name} results: {drop_result}")

    # This should now show empty
    result = execute_query(creds, query, database=database)

    return result

def reset(creds : (str, str, str), database: str = "neo4j"):

    drop_constraints(creds, database)

    # Clears nodes and relationships - but labels remain and can only be cleared via GUI
    query = """MATCH (n) DETACH DELETE n"""
    records, summary, keys = execute_query(creds, query, database=database)

    ModuleLogger().info(f"Reset results: {summary}")
    return summary

def create_new_node_constraints(
        creds: (str, str, str),
        node_key: str,
        database: str = "neo4j"
):
    query = f"""CREATE CONSTRAINT node_key IF NOT EXISTS FOR (u:`{node_key}`)\nREQUIRE u.`node_key` IS UNIQUE"""
    result = execute_query(creds, query, database=database)

    ModuleLogger().info(f"Create new constraints results: {result}")
    return result



from neo4j_uploader.models import GraphData, Nodes, Relationships, TargetNode, Neo4jConfig
from neo4j_uploader._logger import ModuleLogger
from enum import Enum
from copy import deepcopy
import json

class ElementType(Enum):
    UNKNOWN = 0
    NODE = 1
    RELATIONSHIP = 2

def convert_to_hashable(obj):
    if isinstance(obj, dict):
        return tuple({k: convert_to_hashable(v) for k, v in obj.items()}.items())
    elif isinstance(obj, list):
        return tuple(convert_to_hashable(i) for i in obj)
    else:
        return obj
    
def deduped(data: list[any])-> list[any]:
    unique = [] 
    seen = set()

    for d in data:
        # Deepcopy to avoid modifying original object
        tmp = deepcopy(d)
        # Convert nested dicts and lists to tuples for hashing
        tmp = convert_to_hashable(tmp)
        if isinstance(tmp, tuple):
            t = tmp
        else:
            t = tuple(tmp.items())
        
        if t not in seen:
            unique.append(d)
            seen.add(t)
    
    return unique
    

def properties(
        suffix: str,
        record: dict,
        exclude_keys: list[str] = []
    ) -> (str, dict):

    # Sample string output
    # " {`age`:$age_test_0, `name`:$name_test_0}"

    # Sample dict output
    # {
    #   "age_test_0": 30,
    #   "name_test_0": "John Wick"
    # }

    # Convert each batch of records
    result_params = {}
    
    # Filter out unwanted keys
    filtered_keys = [key for key in record.keys() if key not in exclude_keys]

    # Sort keys for consistent testing
    sorted_keys = sorted(list(filtered_keys))


    query = " {"
    for k_idx, a_key in enumerate(sorted_keys):

        value = record[a_key]

        # Do not set properties with a None/Null/Empty value
        if value is None:
            continue
        if isinstance(value, str):
            if value.lower() == "none":
                continue
            if value.lower() == "null":
                continue
            if value.lower() == "empty":
                continue
            if value.lower() == "":
                continue

        # Nested dicts and lists not supported
        if isinstance(value, dict) or isinstance(value, list):
            value = str(value)

        # Prefix multiple items in Cypher with comma
        if k_idx != 0:
            query += ", "

        # Add params for query
        param_key = f'{a_key}_{suffix}'
        result_params[param_key] = value

        # Add string representation of property data
        query += f'`{a_key}`:${param_key}'

    # Close out query
    query += "}"


    return (query, result_params)

# TODO: Update so string portion is formatted as list like relationships
def node_elements(
        batch: str,
        records: list[dict],
        key: str,
        dedupe: bool = True,
        exclude_keys: list[str] = []
    ) -> (str, dict):

    # Sample string output
    # " {`age`:$age_test_0, `name`:$name_test_0}"

    # Sample dict output
    # {
    #   "age_test_0": 30,
    #   "name_test_0": "John Wick"
    # }

    # Remove any duplicates
    # if dedupe == True:
    #     records = [dict(t) for t in {tuple(n.items()) for n in records}]
    if dedupe == True:
        records = deduped(records)

    # Convert each batch of records
    # result_str_list = []
    result_str = ""
    result_params = {}
    for idx, record in enumerate(records):

        # Suffix to uniquely id params
        suffix = f"{batch}{idx}"

        query, param = properties(suffix, record, exclude_keys)
        result_params.update(param)

        key_placeholder = f'{key}_{suffix}'
        key_value = record[key]

        # Nested dicts and lists are not supported
        if isinstance(key_value, dict) or isinstance(key_value, list):
            key_value = str(key_value)

        result_params.update({
            key_placeholder : key_value
        })

        if idx != 0:
            result_str += ", "
        result_str += f"[${key_placeholder}, {query}]"

    # Compile results
    # if len(result_str_list) == 0:
    #     result_str = None
    # else:
    #     result_str = ",".join(result_str_list)
    return (result_str, result_params)

def nodes_query(
        batch: str,
        records: list[dict],
        key: str,
        labels: list[str],
        exclude_keys : list[str] = [],
        dedupe : bool = True
    ) -> (str, dict):
    """Returns a Cypher query for batch uploading node records.

    Args:
        batch (str): String identifier to separate upload batches
        records (list[dict]): List of dictionaries containing Node properties
        labels (list[str]): List of strings designating Node labels
        constraints (list[str], optional): Optional constraints for defining unique Node Property values. Stub for future feature. Defaults to [].
        dedupe (bool, optional): Should duplicates be prevented. True means the Cypher MERGE command will be used. Defaults to True.

    Returns:
        str, dict: Cypher query and params for uploading data.
    """

    # Sample query output
    # WITH [{`uid`:$uid_b0n0, `name`:$name_b0n0},{`uid`:$uid_b0n1, `name`:$name_b1n1}] AS node_data
    # UNWIND node_data as node
    # MERGE (n:`Person` {`uid`:node.`uid`})
    # SET n += node

    # Sample params output
    # {
    #   "uid_test_0":"abc",
    #   "uid_test_1":"cde"
    #   "name_test_0":"John"
    #   "name_test_1":"Cane"
    # }

    if len(records) == 0:
        return None, {}
    
    elements_str, params = node_elements(
        batch = batch,
        records = records,
        key = key,
        dedupe= dedupe,
        exclude_keys= exclude_keys
    )

    if dedupe == True:
        merge_create = "MERGE"
    else:
        merge_create = "CREATE"

    query = f"""WITH [{elements_str}] AS node_data\nUNWIND node_data AS node\n{merge_create} (n:`{labels[0]}` {{ `{key}`:node[0]}} )\nSET n += node[1]"""
    
    if len(labels) > 1:
        for label in labels[1:]:
            query += f'\nSET n:`{label}`'

    return query, params

def relationship_elements(
        batch: str,
        records: list[dict],
        from_node: TargetNode,
        to_node: TargetNode,
        dedupe: bool = True,
        exclude_keys: list[str] = []
    ) -> (str, dict):

   # Sample string output
    # " [$_from__uid_b0r0,$_to__uid_b0r0, {`since`:$since_b0r0}],[...]"

    # Sample dict output
    # {
    #   "_from__uid_b0r0":"123",
    #   "_to__uid_b0r0":"456",
    #   "since_b0r0":2022
    # }

    if len(records) == 0:
        return None, {}
    
    result_str = ""
    result_params = {}

   # Remove any duplicates
    # if dedupe == True:
    #     records = [dict(t) for t in {tuple(n.items()) for n in records}]
    if dedupe == True:
        records = deduped(records)

    for idx, record in enumerate(records):
        
        suffix = f"{batch}{idx}"

        # This is complicated - so we are going to insert the param keys into the string query so that it will be replaced by the value in the params dictionary later to avoid Cypher injection attacks. The same key will be added to the params dict with actual value there. A suffix is necessary to distinguish all the record keys that are aggregated in the in the final params dictionary passed to the Neo4j driver. 
        from_param_key = f"{from_node.record_key}_{suffix}"
        to_param_key = f"{to_node.record_key}_{suffix}"

        props_str, props_params = properties(
            suffix=suffix,
            record=record,
            exclude_keys=exclude_keys
        )

        # Update string
        if idx != 0:
            result_str += ", "
        result_str += f"[${from_param_key}, ${to_param_key},{props_str}]"

        # Update param dict
        result_params.update(props_params)
        result_params.update(
            {
                from_param_key: record[from_node.record_key],
                to_param_key: record[to_node.record_key]
            }
        )

    return result_str, result_params

def relationships_query(
        batch: str,
        records: list[dict],
        from_node: TargetNode,
        to_node: TargetNode,
        type: str,
        exclude_keys : list[str] = [],
        dedupe : bool = True   
    ) -> (str, dict):

    # Sample output
    # WITH [{$from_key_b0r0,$to_key_b0r0, {`a_key`:$a_value}}] AS from_to_data
    # UNWIND from_to_data AS tuple
    # MATCH (fromNode {`$fromKey`:tuple[0]})
    # MATCH (toNode {`$toKey`:tuple[1]})
    # MERGE (fromNode)-[r:`type`]->(toNode)
    # SET r += tuple[2]

    # Sample params output
    # {
    #   "fromKey":"abc",
    #   "uid_test_1":"cde"
    #   "name_test_0":"John"
    #   "name_test_1":"Cane"
    # }

    if len(records) == 0:
        return None, {}
    
    elements_str, params = relationship_elements(
        batch = batch,
        records = records,
        from_node = from_node,
        to_node = to_node,
        dedupe= dedupe,
        exclude_keys= exclude_keys
    )

    # Handle optional Node Label
    from_node_label = from_node.node_label
    to_node_label = to_node.node_label
    if from_node_label == None:
        from_node_label = ""
    else:
        from_node_label = f":`{from_node_label}`"
    if to_node_label == None:
        to_node_label = ""
    else:
        to_node_label = f":`{to_node_label}`"

    # Required Node Key
    from_node_key = f"{from_node.node_key}"
    to_node_key = f"{to_node.node_key}"

    if dedupe == True:
        merge_create = "MERGE"
    else:
        merge_create = "CREATE"

    query = f"""WITH [{elements_str}] AS from_to_data\nUNWIND from_to_data AS tuple\nMATCH (fromNode{from_node_label} {{`{from_node_key}`:tuple[0]}})\nMATCH (toNode{to_node_label} {{`{to_node_key}`:tuple[1]}})\n{merge_create} (fromNode)-[r:`{type}`]->(toNode)\nSET r += tuple[2]"""


    return query, params

def chunked_query(
        spec: Nodes | Relationships,
        config: Neo4jConfig
    ) -> list[(str, dict)]:
    """Returns a list of Cypher queries for batch uploading nodes.

    Args:
        type: 
        records (Any): Nodes or Relationhips model specifying node creation specifications and records
        config (Neo4jConfig): Configuration containing max_batch_size

    Returns:
        list[(str, dict)]: List of queries and params to run for uploading data
    """
    
    # Break up large batches of records
    b = config.max_batch_size
    records = spec.records
    chunked_records = [records[i * b:(i + 1) * b] for i in range((len(records) + b - 1) // b )]  

    # Process each batch into separate query statements
    result = []
    for idx, records in enumerate(chunked_records):
        if isinstance(spec, Nodes):
            query_str, query_params = nodes_query(
                    f"b{idx}n",
                    records,
                    spec.key,
                    spec.labels,
                    spec.exclude_keys,
                    spec.dedupe
                )
        if isinstance(spec, Relationships):

            # Shorthand for automatically excluding keys used to specify source and target nodes
            if spec.auto_exclude_keys is True:
                exclude_keys = [spec.from_node.record_key, spec.to_node.record_key]
            else:
                exclude_keys = spec.exclude_keys
            
            query_str, query_params = relationships_query(
                f"b{idx}r",
                records,
                spec.from_node,
                spec.to_node,
                spec.type,
                exclude_keys,
                spec.dedupe
            )
        if query_str is not None:
            result.append((query_str, query_params))
    return result


def specification_queries(
        specifications: list[Nodes | Relationships],
        config: Neo4jConfig) -> list[(str, dict)]:
    """Returns a list of Cypher queries and params for batch uploading nodes.

    Args:
        specifications (list[Nodes | Relationships]): Nodes and/or Relationships specifications and properties to upload
        config (Neo4jConfig): Configuration containing max_batch_size

    Returns:
        list[(str, dict)]: List of queries and params to run for uploading data
    """

    result = []
    for spec in specifications:
        result.extend(
            chunked_query(
                spec,
                config
            )
        )
    return result


from pydantic import BaseModel, Field
from typing import Optional

# Specify Google doctstring type for pdoc auto doc generation
__docformat__ = "google"

class Neo4jConfig(BaseModel):
    """
    Object for specifying target local or hosted Neo4j database instance to upload to data to.

    Args:
        neo4j_uri (str): The URI of the Neo4j instance to upload to.
        neo4j_password (str): The password for the Neo4j instance to upload to.
        neo4j_user (str): The username for the Neo4j instance to upload to.
        neo4j_database (str): The name of the Neo4j database to upload to. Default 'neo4j'.
        max_batch_size (int): Maximum number of nodes to upload in a single batch. Default 500.
        overwrite (bool): Overwrite existing nodes. Default False.
    """
    neo4j_uri : str
    neo4j_password: str
    neo4j_user : str = Field(default="neo4j")
    neo4j_database : str = Field(default="neo4j")
    max_batch_size : int = Field(default=500)
    overwrite : bool = False

class Nodes(BaseModel):
    """Configuration object for uploading nodes to a Neo4j database.

    Args:
        labels (list[str]): List of node labels to upload (ie Person, Place, etc).
        key (str): Unique key for each node.
        records (list[dict]): List of dictionary objects containing node data.
        exclude_keys (list[str]): List of keys to exclude from upload.
        dedupe (bool): Remove duplicate entries. Default True.
    """
    labels: list[str]
    key: str
    records: list[dict]
    exclude_keys: Optional[list[str]] = []
    dedupe : Optional[bool] = True

class TargetNode(BaseModel):
    """Node specification object for uploading relationships to a Neo4j database.

    Args:
        node_label (str): Optional Node label of the target node. Including a label is more performant than without.
        node_key (str): Target key or property name that identifies a unique node.
        record_key (str): Key within the relationship record, whose value will be mapped to the node_key.
    """
    node_label: Optional[str] = None
    node_key: str
    record_key: str

class Relationships(BaseModel):
    """Configuration object for uploading relationships to a Neo4j database.

    Args:
        type (str): Relationship type (ie FOLLOWS, WORKS_AT, etc).
        from_node (TargetNode): TargetNode object for the source node.
        to_node (TargetNode): TargetNode object for the target node.
        records (list[dict]): List of dictionary objects containing relationship data.
        exclude_keys (list[str]): List of keys to exclude from upload.
        auto_exclude_keys (bool): Automatically exclude keys used to reference nodes used in the from_node and to_node arguments. Default True.
        dedupe (bool): Remove duplicate entries. Default True.
    """
    type: str
    from_node: TargetNode
    to_node: TargetNode
    records : list[dict]
    exclude_keys: Optional[list[str]] = []
    auto_exclude_keys: Optional[bool] = True
    dedupe: Optional[bool] = True

class GraphData(BaseModel):
    """Object representation of nodes and relationships specifications and records to upload to a specified Neo4j database.

    Args:
        nodes (list[Nodes]): List of Nodes configuration objects for specifying nodes to upload.
        relationships (list[Relationships]): Optional List of Relationships configuration objects for specifying relationships to upload.
    """
    nodes: list[Nodes] = []
    relationships: Optional[list[Relationships]] = []

class UploadResult(BaseModel):
    """Result object for uploading nodes to a Neo4j database.

    Args:
        was_successful (bool): True if upload was successful.
        seconds_to_complete (float): Time taken to upload nodes in seconds.
        nodes_created (int): Number of nodes created.
        relationships_created (int): Number of relationships created.
        properties_set (int): Number of properties set.
        error_message (str): Error message if upload failed.
    """
    was_successful : bool
    seconds_to_complete: Optional[float] = None
    nodes_created: Optional[int] = None
    relationships_created: Optional[int] = None
    properties_set: Optional[int] = None
    error_message: Optional[str] = None


# used for converting legacy or alternate schemas to default schema
from neo4j_uploader.models import Nodes, Relationships, TargetNode

def convert_legacy_node_records(
        data: dict,
        dedupe: bool,
        node_key: str
    ) -> list[Nodes]:
    """Convert simple nodes data into new Nodes model object

    Args:
        data (dict): Dictionary containing simple nodes data

    Returns:
        list[Nodes]: List of Nodes objects
    """

    result = []

    # Each key will become a new Nodes specification object
    for key in data.keys():
        node_label = key
        node_key = node_key
        new_node_spec = Nodes(
            labels= [node_label],
            key=node_key,
            records = data[key],
            dedupe=dedupe,
        )
        result.append(new_node_spec)
    
    return result

def convert_legacy_relationship_records(
        data: dict, 
        dedupe: bool,
        node_key: str
    ) -> list[Relationships]:
    """Convert simple relationship data to new Relationships model object

    Args:
        data (dict): Dictionary containing nodes and relationships data

    Returns:
        list[Relationships]: List of Relationships objects
    """

    result = []

    # Each key will become a new Relationships specification object
    for key in data.keys():
        rel_type = key
        from_node = TargetNode(
            node_key = node_key,
            record_key = f"_from_{node_key}",
        )
        to_node = TargetNode(
            node_key = node_key,
            record_key = f"_to_{node_key}",
        )
        records = data[key]

        new_rel_spec = Relationships(
            type = rel_type,
            from_node = from_node,
            to_node = to_node,
            records = records,
            dedupe = dedupe,
        )
        result.append(new_rel_spec)
    
    return result